import pathlib
from shutil import copy2
from typing import List, Tuple
import os
import importlib.util
import sys
import hashlib

import torch
import torch.distributed as dist
from vlutils.base.registry import Registry
from vlutils.config import summary

from mcquic import Config, Consts
from mcquic.modules.compressor import BaseCompressor, Compressor
from mcquic.datasets import getTrainLoader, getValLoader
from mcquic.train.hooks import getAllHooks
from mcquic.utils.registry import *
import mcquic.utils.registry
import mcquic.train.lrSchedulers as _
import mcquic.loss

from .utils import getSaver, initializeBaseConfigs
from .trainer import getTrainer


def registerForTrain(config: Config):
    _registerBuiltinFunctions()

    otherPythonFiles = config.Train.ExternalLib

    _registerExternalFunctions(otherPythonFiles)


def _registerExternalFunctions(otherPythonFiles: List[str]):
    for pyFile in otherPythonFiles:
        filePath = pathlib.Path(pyFile).absolute()
        # md5 of abs file path as module name
        moduleName = hashlib.md5(str(filePath).encode()).hexdigest()
        spec = importlib.util.spec_from_file_location(moduleName, pyFile)
        if spec is None:
            continue
        module = importlib.util.module_from_spec(spec)
        sys.modules[moduleName] = module
        spec.loader.exec_module(module)


def _registerBuiltinFunctions():
    # Built-in pytorch modules to be registered.
    try:
        import apex
        OptimizerRegistry.register("Lamb")(apex.optimizers.FusedLAMB)
    except:
        raise ImportError("`import apex` failed. Apex not installed.")
    OptimizerRegistry.register("Adam")(torch.optim.Adam)

    LrSchedulerRegistry.register("ReduceLROnPlateau")(torch.optim.lr_scheduler.ReduceLROnPlateau)
    LrSchedulerRegistry.register("Exponential")(torch.optim.lr_scheduler.ExponentialLR)
    LrSchedulerRegistry.register("MultiStep")(torch.optim.lr_scheduler.MultiStepLR)
    LrSchedulerRegistry.register("OneCycle")(torch.optim.lr_scheduler.OneCycleLR)


def modelFn(modelParams, lossTarget) -> Tuple[BaseCompressor, mcquic.loss.Distortion]:
    compressor = Compressor(**modelParams)
    criterion = LossRegistry.get(lossTarget)()

    return compressor, criterion


def ddpSpawnTraining(rank: int, worldSize: int, port: str, config: Config, saveDir: str, resume: pathlib.Path, loggingLevel: int):
    registerForTrain(config)


    # load ckpt before create trainer, in case it moved to other place.
    if resume is not None:
        if rank == 0:
            tmpFile = copy2(resume, os.path.join(Consts.TempDir, "resume.ckpt"), follow_symlinks=False)
        else:
            tmpFile = os.path.join(Consts.TempDir, "resume.ckpt")
    else:
        tmpFile = None


    saver = getSaver(saveDir, saveName="saved.ckpt", loggerName=Consts.Name, loggingLevel=loggingLevel, config=config.serialize(), reserve=False, disable=rank != 0)

    saver.info("Here is the whole config during this run: \r\n%s", summary(config.serialize()))

    saver.debug("Creating the world...")

    initializeBaseConfigs(port, rank, worldSize, logger=saver)
    saver.debug("Base configs initialized.")

    dist.barrier()

    for reg in mcquic.utils.registry.__all__:
        registry = getattr(mcquic.utils.registry, reg)
        if issubclass(registry, Registry):
            saver.debug("Summary of %s: \r\n%s", registry, registry.summary())

    optimizerFn = OptimizerRegistry.get(config.Train.Optim.Key, logger=saver)
    schdrFn = LrSchedulerRegistry.get(config.Train.Schdr.Key, logger=saver)

    trainer = getTrainer(rank, config, lambda: modelFn(config.Model.Params, config.Train.Target), optimizerFn, schdrFn, saver)

    if tmpFile is not None:
        saver.info("Found ckpt to resume at %s", resume)
        trainer.restoreStates(tmpFile)

    trainLoader, trainSampler = getTrainLoader(rank, worldSize, config.Train.TrainSet, config.Train.BatchSize, logger=saver)
    valLoader = getValLoader(config.Train.ValSet, disable=rank != 0, logger=saver)
    saver.debug("Train and validation datasets mounted.")

    trainer.train(trainLoader, trainSampler, valLoader, **getAllHooks(config.Train.Hooks))

    saver.debug(summary(config.serialize()))
    saver.info("Bye.")



from typing import Tuple
import random

import torch
from torch import nn
import torch.nn.functional as tf
from torchvision.transforms import functional as F
from torch.distributions import Categorical


__all__ = [
    "RandomPlanckianJitter",
    "RandomGamma",
    "DeTransform",
    "RandomHorizontalFlip",
    "RandomVerticalFlip",
    "RandomAutocontrast",
    "PatchWiseErasing"
]



# BEGIN::: From Kornia
_planckian_coeffs = torch.tensor(
    [
        [0.6743, 0.4029, 0.0013],
        [0.6281, 0.4241, 0.1665],
        [0.5919, 0.4372, 0.2513],
        [0.5623, 0.4457, 0.3154],
        [0.5376, 0.4515, 0.3672],
        [0.5163, 0.4555, 0.4103],
        [0.4979, 0.4584, 0.4468],
        [0.4816, 0.4604, 0.4782],
        [0.4672, 0.4619, 0.5053],
        [0.4542, 0.4630, 0.5289],
        [0.4426, 0.4638, 0.5497],
        [0.4320, 0.4644, 0.5681],
        [0.4223, 0.4648, 0.5844],
        [0.4135, 0.4651, 0.5990],
        [0.4054, 0.4653, 0.6121],
        [0.3980, 0.4654, 0.6239],
        [0.3911, 0.4655, 0.6346],
        [0.3847, 0.4656, 0.6444],
        [0.3787, 0.4656, 0.6532],
        [0.3732, 0.4656, 0.6613],
        [0.3680, 0.4655, 0.6688],
        [0.3632, 0.4655, 0.6756],
        [0.3586, 0.4655, 0.6820],
        [0.3544, 0.4654, 0.6878],
        [0.3503, 0.4653, 0.6933],
        [0.5829, 0.4421, 0.2288],
        [0.5510, 0.4514, 0.2948],
        [0.5246, 0.4576, 0.3488],
        [0.5021, 0.4618, 0.3941],
        [0.4826, 0.4646, 0.4325],
        [0.4654, 0.4667, 0.4654],
        [0.4502, 0.4681, 0.4938],
        [0.4364, 0.4692, 0.5186],
        [0.4240, 0.4700, 0.5403],
        [0.4127, 0.4705, 0.5594],
        [0.4023, 0.4709, 0.5763],
        [0.3928, 0.4713, 0.5914],
        [0.3839, 0.4715, 0.6049],
        [0.3757, 0.4716, 0.6171],
        [0.3681, 0.4717, 0.6281],
        [0.3609, 0.4718, 0.6380],
        [0.3543, 0.4719, 0.6472],
        [0.3480, 0.4719, 0.6555],
        [0.3421, 0.4719, 0.6631],
        [0.3365, 0.4719, 0.6702],
        [0.3313, 0.4719, 0.6766],
        [0.3263, 0.4719, 0.6826],
        [0.3217, 0.4719, 0.6882],
    ]
)

_planckian_coeffs_ratio = torch.stack(
    (
        _planckian_coeffs[:, 0] / _planckian_coeffs[:, 1],
        _planckian_coeffs[:, 2] / _planckian_coeffs[:, 1],
    ),
    1,
)

class RandomPlanckianJitter(nn.Module):
    pl: torch.Tensor
    def __init__(self, p: float = 0.5) -> None:
        super().__init__()
        self.register_buffer('pl', _planckian_coeffs_ratio)
        self.p = p

    def forward(self, x: torch.Tensor):
        needsApply = torch.rand(x.shape[0]) < self.p
        coeffs = self.pl[torch.randint(len(self.pl), (needsApply.sum(), ))]

        r_w = coeffs[:, 0][..., None, None]
        b_w = coeffs[:, 1][..., None, None]

        willbeApplied = x[needsApply]

        willbeApplied[..., 0, :, :].mul_(r_w)
        willbeApplied[..., 2, :, :].mul_(b_w)

        return x.clamp_(0.0, 1.0)



def srgbToLinear(x: torch.Tensor):
    return torch.where(x < 0.0031308, 12.92 * x, (1.055 * torch.pow(torch.abs(x), 1 / 2.4) - 0.055))

def linearToSrgb(x: torch.Tensor):
    return torch.where(x < 0.04045, x / 12.92, torch.pow(torch.abs(x + 0.055) / 1.055, 2.4))

def randomGamma(x: torch.Tensor, randomGammas: torch.Tensor):
    x = torch.pow(x.clamp_(0), randomGammas)
    return x.clamp_(0.0, 1.0)

def identity(x: torch.Tensor):
    return x

class RandomGamma(nn.Module):
    _fns = [
        srgbToLinear,
        linearToSrgb,
        lambda x: randomGamma(x, torch.rand((), device=x.device) * 1.95 + 0.05),
        lambda x: x
    ]
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return random.choice(self._fns)(x)


# END::: From Kornia


class DeTransform(nn.Module):
    _eps = 1e-3
    _maxVal = 255
    def __init__(self, minValue: float = -1.0, maxValue: float = 1.0):
        super().__init__()
        self._min = float(minValue)
        self._max = float(maxValue)

    def forward(self, x):
        x = (x - self._min) / (self._max - self._min)
        # [0, 1] to [0, 255]
        return (x * (self._maxVal + 1.0 - self._eps)).clamp(0.0, 255.0).byte()


# https://github.com/pratogab/batch-transforms
class RandomHorizontalFlip(nn.Module):
    """Applies the :class:`~torchvision.transforms.RandomHorizontalFlip` transform to a batch of images.
    Args:
        p (float): probability of an image being flipped.
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor of size (N, C, H, W) to be flipped.
        Returns:
            Tensor: Randomly flipped Tensor.
        """
        flipped = torch.rand(tensor.shape[0]) < self.p
        tensor[flipped].copy_(tensor[flipped].flip(-1))
        return tensor

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomVerticalFlip(nn.Module):
    """Applies the :class:`~torchvision.transforms.RandomVerticalFlip` transform to a batch of images.
    Args:
        p (float): probability of an image being flipped.
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor of size (N, C, H, W) to be flipped.
        Returns:
            Tensor: Randomly flipped Tensor.
        """
        flipped = torch.rand(tensor.shape[0]) < self.p
        tensor[flipped].copy_(tensor[flipped].flip(-2))
        return tensor

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomAutocontrast(torch.nn.Module):
    """Autocontrast the pixels of the given image randomly with a given probability.
    If the image is torch Tensor, it is expected
    to have [..., 1 or 3, H, W] shape, where ... means an arbitrary number of leading dimensions.
    If img is PIL Image, it is expected to be in mode "L" or "RGB".
    Args:
        p (float): probability of the image being autocontrasted. Default value is 0.5
    """

    def __init__(self, p=0.5):
        super().__init__()
        self.p = p

    def forward(self, img):
        """
        Args:
            img (PIL Image or Tensor): Image to be autocontrasted.
        Returns:
            PIL Image or Tensor: Randomly autocontrasted image.
        """
        picked = torch.rand(img.shape[0]) < self.p
        img[picked].copy_(F.autocontrast(img[picked]))
        return img

    def __repr__(self):
        return self.__class__.__name__ + f"(p={self.p})"


class Masking(nn.Module):
    def __init__(self):
        super().__init__()
        self._categorical = Categorical(torch.Tensor([0.85, 0.15]))

    def forward(self, images: torch.Tensor):
        n, _, h, w = images.shape
        zeros = torch.zeros_like(images)
        # [n, 1, h, w]
        mask = self._categorical.sample((n, 1, h, w)).byte().to(images.device)
        return (mask == 0) * images + (mask == 1) * zeros


class PatchWiseErasing(nn.Module):
    permutePattern: torch.Tensor
    def __init__(self, grids: Tuple[int, int] = (16, 16), p: float = 0.75) -> None:
        super().__init__()
        self.p = p
        self.grids = (1, 1, *grids)
        # [1024, ]
        permutePattern = torch.ones(grids).flatten()
        permuteAmount = round(permutePattern.numel() * p)
        permutePattern[:permuteAmount] = 0
        self.register_buffer("permutePattern", permutePattern)

    def forward(self, x: torch.Tensor):
        shape = x.shape
        h, w = shape[-2], shape[-1]
        randIdx = torch.randperm(len(self.permutePattern), device=x.device)
        self.permutePattern.copy_(self.permutePattern[randIdx])
        permutePattern = self.permutePattern.reshape(self.grids)
        eraseArea = tf.interpolate(permutePattern, (h, w))
        return x * eraseArea



import abc
from enum import Enum
from functools import wraps
from typing import Any, Callable, List, Union, Dict

from vlutils.saver import Saver
from vlutils.base.freqHook import ChainHook
from vlutils.base.registry import Registry
from mcquic.config import General
from mcquic.train.utils import getSaver

from mcquic.utils.registry import HookRegistry
from mcquic.train.trainer import _baseTrainer


class HookType(Enum):
    BeforeRunHook = "beforeRunHook"
    AfterRunHook = "afterRunHook"
    EpochStartHook = "epochStartHook"
    EpochFinishHook = "epochFinishHook"
    StepStartHook = "stepStartHook"
    StepFinishHook = "stepFinishHook"

    def __str__(self):
        return str(self.value)


def hook(hookType: HookType):
    """A decorator that marks a callable as a hook.

    NOTE: Please notice the call signature of hooks must follow example given in Usage.

    Usage:

    ```python
        # To call the hook, the function registered should follow signature below
        # where `step, epcoh` is total train steps and epochs. `trainer` is the trainer instance.
        # All hooks are called after trainer's original methods.
        # It's not recommended to modify trainer in hooks (Only if you really need indeed).
        # Some other arguments may be passed in. Please refer to `mcquic/train/trainer.py:L123` to see what are passed.
        @hook(HookType.xxxhook)
        def someFunction(step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
            ...

        # For classes, since decorated function is not a instance's method, you should use `@classmethod`.
        # However, this is not recommended and has limitations.
        # Inherit from `XXXHook` is better, which is defined below.
        class SomeClass:
            ...
            @hook(HookType.xxxhook)
            @classmethod
            def someFunction(cls, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
                ...
    ```

    Args:
        hookType (HookType): Hook type.
    """
    def _hook(fn: Callable):
        @wraps(fn)
        def _call(*args, **kwargs):
            return fn(*args, **kwargs)
        _call.hookType = hookType
        return _call
    return _hook

"""
Implement hooks by inheriting from one or multiple hook ABCs.
Please see built-in hooks below as examples.
"""
class BeforeRunHook(abc.ABC):
    @abc.abstractmethod
    def beforeRun(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        raise NotImplementedError
class AfterRunHook(abc.ABC):
    @abc.abstractmethod
    def afterRun(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        raise NotImplementedError
class EpochStartHook(abc.ABC):
    @abc.abstractmethod
    def epochStart(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        raise NotImplementedError
class EpochFinishHook(abc.ABC):
    @abc.abstractmethod
    def epochFinish(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        raise NotImplementedError
class StepStartHook(abc.ABC):
    @abc.abstractmethod
    def stepStart(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        raise NotImplementedError
class StepFinishHook(abc.ABC):
    @abc.abstractmethod
    def stepFinish(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        raise NotImplementedError


# Some built-in hooks START
@HookRegistry.register
class DisablePostProcessAfterEpoch(EpochStartHook):
    def __init__(self, epoch: int) -> None:
        super().__init__()
        self._epoch = epoch

    def epochStart(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        if epoch > self._epoch > 0:
            trainer._model.PostProcessEnabled = False
            logger.debug(f"Set PostProcessEnabled to `False` at epoch {epoch}.")
        else:
            trainer._model.PostProcessEnabled = True
            logger.debug(f"Set PostProcessEnabled to `True` at epoch {epoch}.")

@HookRegistry.register
class CodebookReassign(EpochFinishHook):
    def __init__(self, freq) -> None:
        super().__init__()
        self._freq = freq

    def epochFinish(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, logger: Saver, **kwds: Any) -> Any:
        if epoch % self._freq != 0:
            return
        logger.debug("[%s] Start refresh at epoch %4d.", trainer.PrettyStep, epoch)

        reAssignProportion = trainer._model.refresh(trainer.rank)

        logger.debug("[%s] %.2f%% of codebook is re-assigned.", trainer.PrettyStep, reAssignProportion * 100)

        logger.debug("[%s] End refresh at epoch %4d.", trainer.PrettyStep, epoch)

        logger.add_scalar("Stat/ReAssignProportion", reAssignProportion, global_step=step)



class TrainerLogger(BeforeRunHook, AfterRunHook, EpochStartHook, EpochFinishHook, StepStartHook, StepFinishHook):
    def __init__(self):
        super().__init__()

    def beforeRun(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, **kwds: Any) -> Any:
        return

    def afterRun(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, **kwds: Any) -> Any:
        return

    def epochStart(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, **kwds: Any) -> Any:
        return

    def epochFinish(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, **kwds: Any) -> Any:
        return

    def stepStart(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, **kwds: Any) -> Any:
        return

    def stepFinish(self, step: int, epoch: int, trainer: _baseTrainer, *args: Any, **kwds: Any) -> Any:
        return
# Some built-in hooks END




def splitHooks(*hooks: Union[Callable, BeforeRunHook, AfterRunHook, EpochStartHook, EpochFinishHook, StepStartHook, StepFinishHook]) -> Dict[HookType, ChainHook]:
    """Split hooks into beforeRunHook, afterRunHook, epochStartHook, epochFinishHook, stepStartHook, stepFinishHook.

    Args:
        hooks (List[Callable]): Hooks to be split.

    Returns:
        Tuple[ChainHook, ...]: Tuple of hooks.
    """
    allHooks = { k: list() for k in HookType }

    hookToTypeMap = {
        BeforeRunHook: HookType.BeforeRunHook,
        AfterRunHook: HookType.AfterRunHook,
        EpochStartHook: HookType.EpochStartHook,
        EpochFinishHook: HookType.EpochFinishHook,
        StepStartHook: HookType.StepStartHook,
        StepFinishHook: HookType.StepFinishHook
    }

    hookToCallableMap = {
        BeforeRunHook: "beforeRun",
        AfterRunHook: "afterRun",
        EpochStartHook: "epochStart",
        EpochFinishHook: "epochFinish",
        StepStartHook: "stepStart",
        StepFinishHook: "stepFinish"
    }

    for hook in hooks:
        if isinstance(hook, (BeforeRunHook, AfterRunHook, EpochStartHook, EpochFinishHook, StepStartHook, StepFinishHook)):
            for hookClass, hookType in hookToTypeMap.items():
                if isinstance(hook, hookClass):
                    allHooks[hookType].append(getattr(hook, hookToCallableMap[hookClass]))
        elif hasattr(hook, "hookType"):
            allHooks[hook.hookType].append(hook)
        else:
            raise ValueError(f"Unknown hook type of given value `{hook}`.")

    return {
        k: ChainHook(*v) for k, v in allHooks.items()
    }


def getAllBuiltinHooks() -> Dict[HookType, ChainHook]:
    allHooks = list()
    for hook in BuiltInHooks.values():
        if hasattr(hook, "hookType"):
            # A decorated function. Append it directly.
            allHooks.append(hook)
        else:
            # A class-def, create it.
            allHooks.append(hook())
    return splitHooks(*allHooks)


def getAllHooks(otherHooks: List[General]) -> Dict[str, ChainHook]:
    builtInHooks = getAllBuiltinHooks()

    otherHooksToAppend = list()
    for hook in otherHooks:
        hookFn = HookRegistry.get(hook.Key)
        if hasattr(hookFn, "hookType"):
            # A decorated function. Append it directly.
            otherHooksToAppend.append(hookFn)
        else:
            # A class-def, create it.
            otherHooksToAppend.append(hookFn(**hook.Params))
    otherHooksToAppend = splitHooks(*otherHooksToAppend)

    allHooks: Dict[str, ChainHook] = dict()
    for key in builtInHooks.keys():
        allHooks[str(key)] = ChainHook(builtInHooks[key], otherHooksToAppend[key])
    return allHooks



from typing import Optional, Any, Union, Sequence

from torch import device, Tensor
import torch
from torch.nn import Module
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel

from mcquic.loss import Distortion
from mcquic.datasets.transforms import getTraingingPostprocess
from .compressor import BaseCompressor

_device_t = Union[int, device]
_devices_t = Sequence[_device_t]


class _compound(Module):
    def __init__(self, compressor: BaseCompressor, criterion: Distortion):
        super().__init__()
        self._compressor = compressor
        self._criterion = criterion
        self._postProcess = getTraingingPostprocess().to(dist.get_rank())
        self._postProcessEnabled = True

    def forward(self, x: Tensor):
        if self._postProcessEnabled:
            post = self._postProcess(x)
        else:
            post = x
        xHat, yHat, codes, logits = self._compressor(post)
        rate, distortion = self._criterion(x, xHat, codes, logits)
        return (post, xHat), (rate, distortion), codes, logits

    @property
    def Freq(self):
        return self._compressor._quantizer._entropyCoder.NormalizedFreq

    @property
    def PostProcessEnabled(self):
        return self._postProcessEnabled

    @PostProcessEnabled.setter
    def PostProcessEnabled(self, enabled: bool):
        self._postProcessEnabled = enabled


class Compound(DistributedDataParallel):
    module: _compound
    def __init__(self, compressor: BaseCompressor, criterion: Distortion, device_ids: Optional[_devices_t] = None, output_device: Optional[_device_t] = None, dim: int = 0, broadcast_buffers: bool = True, process_group: Optional[Any] = None, bucket_cap_mb: float = 25, find_unused_parameters: bool = False, **kwargs):
        module = _compound(compressor, criterion)
        super().__init__(module, device_ids, output_device, dim, broadcast_buffers, process_group, bucket_cap_mb, find_unused_parameters, **kwargs)

    @property
    def Compressor(self):
        return self.module._compressor

    def refresh(self, rank: int) -> torch.Tensor:
        if rank == 0:
            proportion = self.Compressor.reAssignCodebook()
        else:
            proportion = torch.zeros(())
        self.Compressor.syncCodebook()
        return proportion

    def formatDistortion(self, loss: torch.Tensor):
        return self.module._criterion.formatDistortion(loss)

    @property
    def PostProcessEnabled(self):
        return self.module.PostProcessEnabled

    @PostProcessEnabled.setter
    def PostProcessEnabled(self, enabled: bool):
        self.module.PostProcessEnabled = enabled
